from pydantic import BaseModel
from typing import Dict
import requests, json, time, os
from datetime import datetime

class Property(BaseModel):
    type: str
    description: str

class InputSchema(BaseModel):
    type: str
    properties: Dict[str, Property]
    required: list

class AnthropicTool(BaseModel):
    name: str
    description: str
    input_schema: InputSchema

class OpenAITool(BaseModel):
    name: str
    description: str
    parameters: InputSchema

class AnthropicRetrieverTool:
    @staticmethod
    def create_tool() -> AnthropicTool:
        input_schema = InputSchema(
            type="object",
            properties={
                "query": Property(
                    type="string",
                    description="The user query to be used for retrieval. The query must be comprehensive and in Thai"
                )
            },
            required=["query"]
        )
        return AnthropicTool(
            name="retriever",
            description="Retrieve documents related to a given query from a vector database",
            input_schema=input_schema
        )

class OpenAIRetrieverTool:
    @staticmethod
    def create_tool() -> OpenAITool:
        parameters = InputSchema(
            type="object",
            properties={
                "query": Property(
                    type="string",
                    description="The user query to be used for retrieval. The query must be comprehensive and in Thai"
                )
            },
            required=["query"]
        )
        return OpenAITool(
            name="retriever",
            description="Retrieve documents related to a given query from a vector database",
            parameters=parameters
        )


class MCPTools:
    """MCP (Model Context Protocol) tool execution client.

    Connects to an MCP server to discover available tools and execute them.

    Args:
        mcp_url: Base URL of the MCP server
        user_chat_id: User/tenant chat ID for multi-tenant isolation
        user_context: User context dict passed to tool executions
        mcp_api_key: API key for MCP server. Defaults to MCP_API_KEY env var.
    """
    def __init__(self, mcp_url: str, user_chat_id: str, user_context: Dict = None, mcp_api_key: str = None, profiler=None):
        self.mcp_url = mcp_url
        self.user_chat_id = user_chat_id
        self.mcp_api_key = mcp_api_key or os.environ.get("MCP_API_KEY", "")
        self.profiler = profiler
        if self.profiler:
            self.profiler.mark("mcp_list_tools_start")
        self.tools = self._retrieve_tools_list()
        if self.profiler:
            self.profiler.mark("mcp_list_tools_end")
            self.profiler.record_duration("mcp_list_tools", self.profiler.between("mcp_list_tools_start", "mcp_list_tools_end"))
        self.user_context = user_context or {}

    def _retrieve_tools_list(self):
        res = requests.post(
            f"{self.mcp_url}/list_tools",
            headers={"x-api-key": self.mcp_api_key},
            json={"userChatID": self.user_chat_id}
        )
        if res.status_code != 200:
            raise ConnectionError(f"MCP list_tools failed (HTTP {res.status_code}): {res.text[:200]}")
        body = res.json()
        if "availableTools" not in body:
            raise KeyError(f"MCP response missing 'availableTools'. Got: {list(body.keys())}")
        tools = body["availableTools"]
        has_rag = any(t["name"] == "rag" for t in tools)
        if not has_rag:
            rag = {
                "type": "function",
                "name": "rag",
                "description": "RAG tool for semantic search",
                "endpoint": {"method": "POST", "path": "/rag"},
                "parameters": {
                    "type": "object",
                    "properties": {"query": {"type": "string"}},
                    "required": ["query"]
                }
            }
            tools.append(rag)
        return tools

    def execute_tool(self, tool_name: str, user_message=None, **data):
        if "userChatID" not in data:
            data["userChatID"] = self.user_chat_id
        if "dateTime" not in data:
            data["dateTime"] = datetime.now().strftime('%d/%m/%Y %H:%M:%S')
        if "userContext" not in data:
            data["userContext"] = self.user_context
        if tool_name == "send_email" and user_message:
            data["userMessage"] = user_message
        if tool_name == "rag":
            data["useBm25"] = True

        start = time.perf_counter()
        res = requests.post(
            f"{self.mcp_url}/{tool_name}",
            headers={"x-api-key": self.mcp_api_key},
            json=data
        )
        duration = time.perf_counter() - start

        if self.profiler:
            self.profiler.count("tool_calls")
            self.profiler.record_duration("mcp_tool_exec", duration)
            self.profiler.record_duration(f"mcp_{tool_name}", duration)

        return res.json()

    @staticmethod
    def parse_response(tool_name: str, response: dict, tool_input):
        metadata = None
        scores = None
        if tool_name == "rag":
            if response["status"] == "failed":
                return response["result"], None

            output, metadata = response["output"], response["metadata"]
            if isinstance(metadata, str):
                metadata = json.loads(metadata)
            scores = [item["score"] for item in metadata["documents"]]
        else:
            output = response["result"]
            metadata = None

        print(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": f"Tool {tool_name} results. input: {str(tool_input)} result: <result>{output[:500]}</result> scores: {str(scores)}",
            "exception": None
        }, ensure_ascii=False))

        return output, metadata


if __name__ == "__main__":
    print(AnthropicRetrieverTool.create_tool().model_dump())


from typing import Dict, Tuple, List
from collections import defaultdict
import asyncio, time, json, uuid
from fastapi import BackgroundTasks
from animuz_core.utils import init_logger

import anthropic

from animuz_core.genai.base import BaseLLM, BaseAgent
from animuz_core.genai.prompts import CLAUDE_RAG_DEFAULT_PROMPT, CLAUDE_TOOL_DEFAULT_PROMPT, SUMMARIZE_PROMPT, COMBINE_SUMMARIES_PROMPT
from animuz_core.genai.tools import AnthropicTool

LOGGER = init_logger(__name__)

class AnthropicClient(BaseLLM):
    def __init__(self, system_prompt: str = None) -> None:
        self.client = anthropic.AsyncClient()
        self.system_prompt = CLAUDE_RAG_DEFAULT_PROMPT if system_prompt is None else system_prompt
        self.message_history = []
        self.temperature = 0.0
        self.model = "claude-3-haiku-20240307"
        self.max_token = 1000
        self.semaphore = asyncio.Semaphore(2)

    async def __aenter__(self):
        self.client_context = await self.client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.client_context.__aexit__(exc_type, exc_value, traceback)

    async def get_reply(self, text: str) -> str:
        """
        Appends the user's input text to the message history, creates a message using the client with specified parameters, 
        and appends the assistant's response to the message history. Returns the assistant's response text.
        
        Parameters:
            text (str): The user's input text.
        
        Returns:
            str: The assistant's response text.
        """
        self.message_history.append({"role": "user", "content": text})
      
        message = await self.client.messages.create(
            model=self.model,
            max_tokens=self.max_token,
            temperature=self.temperature,
            system=self.system_prompt,
            messages=self.message_history
        )
        
        self.message_history.append({"role": "assistant", "content": message.content[-1].text})
        return message.content[-1].text
    

class AnthropicAgentClient(BaseAgent):
    def __init__(self, system_prompt: str = None, tools: Dict[str, Tuple[AnthropicTool, callable]] = None) -> None:
        self.client = anthropic.AsyncClient()
        self.system_prompt = CLAUDE_TOOL_DEFAULT_PROMPT if system_prompt is None else system_prompt
        self.temperature = 0.0
        self.model = "claude-3-haiku-20240307"
        self.max_token = 1000
        if tools:
            self.add_tools(tools)

        self.result_store = {}

    def add_tools(self, tools: Dict[str, Tuple[AnthropicTool, callable]]):
        self.tools = tools
        self.tool_dump = [tool[0].model_dump() for tool in self.tools.values()]

    def _filter_for_anthropic(self, messages: List[Dict]) -> List[Dict]:
        """Remove non-Anthropic fields (metadata) before sending to API."""
        clean = []
        for msg in messages:
            m = msg.copy()
            m.pop("metadata", None)
            clean.append(m)
        return clean

    async def get_reply_frontend(self, messages: List[Dict[str, str]], user_chat_id: str, system_prompt: str = None) -> str:
        """
        Asynchronously retrieves a reply from the Anthropic Agent using the provided messages and system prompt.

        Args:
            messages (List[Dict[str, str]]): A list of dictionaries representing the messages in the conversation.
            system_prompt (str): The system prompt to use when creating the message.
            collection_id (str): The ID of the collection to use when using the retrieval tool.

        Returns:
            str: The content of the last message in the conversation.

        This function creates a message using the provided messages and system prompt, and appends the assistant's response to the messages list.
        If the message stop reason is not "tool_use", the assistant's response is appended to the output list.
        If the message stop reason is "tool_use", the function enters a loop where it processes the tool use response,
        appends the tool result to the messages and output lists, and creates a new message.
        The assistant's response for each new message is appended to the messages and output lists.
        The function returns the content of the last message in the conversation.
        """

        message = await self.create_message(messages, system_prompt)
        output = []
        if message.stop_reason != "tool_use":
           messages.append({"role": "assistant", "content": message.content[-1].text})
           output.append({"role": "assistant", "content": message.content[-1].text})
        else:
            messages.append({"role": "assistant", "content": message.content})
            output.append({"role": "assistant", "content": message.content})

        while message.stop_reason == "tool_use":
            tool_use_response = message.content[-1]
            result, metadata = await self.process_tool_use(tool_use_response.input, tool_use_response.name, user_chat_id)

            tool_result = {
                "role": "user",
                "content": [{"type": "tool_result", "tool_use_id": tool_use_response.id, "content": result, "is_error": result == 'ERROR'}],
                "metadata": metadata
            }
            messages.append(tool_result)
            output.append(tool_result)

            message = await self.create_message(messages, system_prompt)
            messages.append({"role": "assistant", "content": message.content[-1].text})
            output.append({"role": "assistant", "content": message.content[-1].text})

        return output

    async def get_reply_frontend_tool_use_ack(self, messages: List[Dict[str, str]], user_chat_id: str, background_tasks: BackgroundTasks, system_prompt: str = None) -> str:
        message = await self.create_message(messages, system_prompt)
        output = []

        if message.stop_reason != "tool_use":
            messages.append({"role": "assistant", "content": message.content[-1].text})
            output.append({"role": "assistant", "content": message.content[-1].text})
            return output, None
        else:
            messages.append({"role": "assistant", "content": message.content})
            output.append({"role": "assistant", "content": message.content})
            chat_id = str(uuid.uuid4()) 
            background_tasks.add_task(self.process_full_response, message, messages, user_chat_id, chat_id)

            return output, chat_id

    async def process_full_response(self, message, messages: List[Dict[str, str]], user_chat_id: str, chat_id: str, system_prompt: str = None) -> str:
        output = []

        while message.stop_reason == "tool_use":
            tool_use_response = message.content[-1]
            LOGGER.debug("waiting for tool")
            result, metadata = await self.process_tool_use(tool_use_response.input, tool_use_response.name, user_chat_id)

            tool_result = {
                "role": "user",
                "content": [{"type": "tool_result", "tool_use_id": tool_use_response.id, "content": result, "is_error": result == 'ERROR'}],
                "metadata": metadata
            }
            messages.append(tool_result)
            output.append(tool_result)
            
            message = await self.create_message(messages, system_prompt)
            messages.append({"role": "assistant", "content": message.content[-1].text})
            output.append({"role": "assistant", "content": message.content[-1].text})

        self.result_store[chat_id] = output

    async def get_chat_result(self, chat_id: str):
        if chat_id in self.result_store:
            result = self.result_store[chat_id]
            del self.result_store[chat_id]  # Clean up the stored result
            return result
        else:
            return "Result not ready yet"

    async def create_message(self, history: List[Dict] = None, system_prompt: str = None):
        start = time.time()
        res = await self.client.messages.create(
            model=self.model,
            max_tokens=self.max_token,
            temperature=self.temperature,
            system=self.system_prompt if system_prompt is None else system_prompt,
            messages=self._filter_for_anthropic(self.message_history if history is None else history),
            tools=self.tool_dump
        )

        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Anthropic API latency",
            "latency": round(time.time() - start, 4),
            "exception": None
        }))
        return res

    async def stream_reply_frontend(self, messages: List[Dict[str, str]], system_prompt: str = None):
        current_tool_call = None
        tool_input_buffer = ""

        async with self.client.messages.stream(
            model=self.model,
            max_tokens=self.max_token,
            temperature=self.temperature,
            system=self.system_prompt if system_prompt is None else system_prompt,
            messages=self._filter_for_anthropic(messages),
            tools=self.tool_dump
        ) as stream:
            async for chunk in stream:
                if chunk.type == "content_block_start":
                    if chunk.content_block.type == "tool_use":
                        current_tool_call = {"id": chunk.content_block.id, "name": chunk.content_block.name} 
                        tool_input_buffer = ""
                        yield {"type": "tool_call_start", "content": {"tool_name": current_tool_call["name"]}}
                    else:
                        # start of text
                        yield {"type": "content_start", "content": ""}

                elif chunk.type == "content_block_delta":
                    # Streaming text content
                    if current_tool_call:
                        tool_input_buffer += chunk.delta.partial_json
                        yield {"type": "content", "content": chunk.delta.partial_json}
                    else:
                        yield {"type": "content", "content": chunk.delta.text}
                 
                elif chunk.type == "content_block_stop":
                    # End of content block
                    if current_tool_call:
                        yield {
                            "type": "tool_call_end",
                            "content": chunk.content_block,
                        }
                        current_tool_call = None
                        tool_input_buffer = ""

                    else:
                        yield {"type": "content_end", "content": ""}
                elif chunk.type == "message_stop":
                    # End of message
                    yield {"type": "message_end", "content": ""}
            
    async def get_reply_frontend_streaming(self, messages: List[Dict[str, str]], user_chat_id: str, system_prompt: str = None):
        output = []
        current_message = {"role": "assistant", "content": ""}
        tool_use = False
        async for chunk in self.stream_reply_frontend(messages, system_prompt):
            if chunk["type"] == "content":
                current_message["content"] += chunk["content"]
                yield chunk
            elif chunk["type"] == "content_start":
                yield chunk
            elif chunk["type"] == "content_end":
                yield chunk
            elif chunk["type"] == "tool_call_start":
                # current_message["content"] += chunk["content"]
                yield chunk
            elif chunk["type"] == "tool_call_end":
                chunk["content"] = chunk["content"].to_dict()
                yield chunk # yield tool call message

                # use tool
                tool_use = True
                result, metadata = await self.process_tool_use(chunk["content"]["input"], chunk["content"]["name"], user_chat_id)
                tool_result = {
                    "role": "user",
                    "content": [{
                        "type": "tool_result",
                        "tool_use_id": chunk["content"]["id"],
                        "content": result,
                        "is_error": result == "ERROR"
                    }],
                    "metadata": metadata
                }
                output.append(tool_result)
                messages.append({"role": "assistant", "content": [chunk["content"]]}) # add tool call
                messages.append(tool_result) # add tool result

                yield {"type": "tool_result", "content": tool_result} # yield tool result message
            elif chunk["type"] == "message_end":
                if current_message["content"]:
                    output.append(current_message)
                yield {"type": "end", "content": output}

        while tool_use:
            current_message = {"role": "assistant", "content": ""}
            tool_use = False
            async for chunk in self.stream_reply_frontend(messages, user_chat_id, system_prompt):
                if chunk["type"] == "content":
                    current_message["content"] += chunk["content"]
                    yield chunk
                elif chunk["type"] == "content_start":
                    yield chunk
                elif chunk["type"] == "content_end":
                    yield chunk
                elif chunk["type"] == "tool_call_start":
                    current_message["content"] += chunk["content"]
                    yield chunk
                elif chunk["type"] == "tool_call_end":
                    chunk["content"] = chunk["content"].to_dict()
                    yield chunk # yield tool call message

                    tool_use = True
                    result, metadata = await self.process_tool_use(chunk["content"]["input"], chunk["content"]["name"], user_chat_id)
                    tool_result = {
                        "role": "user",
                        "content": [{
                            "type": "tool_result",
                            "tool_use_id": chunk["content"]["id"],
                            "content": result,
                            "is_error": result == "ERROR"
                        }],
                        "metadata": metadata
                    }
                    output.append(tool_result)
                    messages.append({"role": "assistant", "content": [chunk["content"].to_dict()]}) # add tool call
                    messages.append(tool_result) # add tool result

                    yield {"type": "tool_result", "content": tool_result} # yield tool result message
                elif chunk["type"] == "message_end":
                    if current_message["content"]:
                        output.append(current_message)
                    yield {"type": "end", "content": output}

        if output:
            yield {"type": "final_output", "content": output}



import os, asyncio
from abc import ABC, abstractmethod
from typing import List, Dict, Tuple

from aiobotocore.session import get_session

from qdrant_client import models

from animuz_core.embedding.embedding_client import EmbeddingClient
from animuz_core.embedding.s3_embedding_client import S3EmbeddingClient
from animuz_core.ingest.custom_unstructured_client import MyUnstructuredClient
from animuz_core.ingest.azure_doc_ai_client import AzureDocAiClient
from animuz_core.vectordb.utils import MyQdrantClient
from animuz_core.ingest.structured import Structured
    
class ResponseGetChunk:
    text: str
    source: str

def get_file_extension(key: str) -> str:
    parts = key.split('.')
    return parts[-1]

class BasePipeline(ABC):
    def __init__(self, embedding_client: EmbeddingClient, 
                 db_client: MyQdrantClient, 
                 unstructured_client: MyUnstructuredClient = None, 
                 azure_doc_ai_client: AzureDocAiClient = None
                 ) -> None:
        
        self.embedding_client = embedding_client
        self.db_client = db_client

        self.unstructured_client = unstructured_client if unstructured_client else MyUnstructuredClient(host=os.environ.get("UNSTRUCTURED_HOST"), port=os.environ.get("UNSTRUCTURED_PORT"))
        self.azure_doc_ai_client = azure_doc_ai_client if azure_doc_ai_client else AzureDocAiClient(endpoint=os.environ.get("AZURE_DOC_AI_ENDPOINT"), key=os.environ.get("AZURE_DOC_AI_KEY"))
        
    async def parse_doc(self, path: str) -> List[dict]:

        assert os.path.isfile(path), "Must be a valid file"

        if not path.endswith(".pdf"):
            res = await self.unstructured_client.a_partition(path)
        else:
            res = await self.azure_doc_ai_client.analyze_document(path) #TODO: maybe find an open source for this

        return res
    
    async def parse_doc_markdown(self, path: str, is_chunked: bool) -> Tuple[List[ResponseGetChunk], int]:
        print('========== START PARSE DOC MARKDOWN ========== ')
        print('Path => ', path)
        assert os.path.isfile(path), "Must be a valid file"

        with open(path, "rb") as f:
            buffer = f.read()

        file_extension = get_file_extension(path)

        if is_chunked:
            return Structured.get_chunks(buffer)

        if file_extension in ['txt', 'md', 'tex', 'csv']:
            return Structured.split_file(buffer, file_extension)
        elif file_extension in ['pdf']:
            # The charcount of pdf will be chunked counts instead of file counts (but shouldn't be used)
            res = await self.azure_doc_ai_client.analyze_document(path)
            return res, sum([len(chunk.text) for chunk in res])
        else:
            # The charcount of pdf will be chunked counts instead of file counts (but shouldn't be used)
            res = await self.unstructured_client.a_partition(path)
            return res, sum([len(chunk.text) for chunk in res])


    async def batch_embed(self, texts: List[str]) -> List[dict]:
        return await self.embedding_client.batch_get_embedding(texts)
    
    async def upload_to_db(self, dense_embeddings: List[List[float]], sparse_indices: List[List[int]], sparse_values: List[List[float]], properties: List[Dict[str, str]]) -> None:
        await self.db_client.upload_batch(dense_embeddings, sparse_indices, sparse_values, properties)
    
    async def upload_to_db_v2(self, dense_embeddings: List[List[float]], list_of_sparse_vectors: List[Dict[str, Dict[str, List]]], properties: List[Dict[str, str]]) -> None:
        await self.db_client.upload_batch_v2(dense_embeddings, list_of_sparse_vectors, properties)

    async def embed_single(self, query: str) -> Dict[str, float]:
        return await self.embedding_client.get_embedding(query)
    
    async def query_db(self, query: str, user_chat_id: str, hybrid: bool = True, top_k: int = 3) -> Tuple[List[str], List[models.ScoredPoint]]:
        emb_res = await self.embed_single(query)

        indices = emb_res["sparse_embedding"].keys()
        values = emb_res["sparse_embedding"].values()

        if hybrid:
            search_results = await self.db_client.hybrid_search(emb_res["embedding"], indices, values, user_chat_id, top_k=top_k)
            return [res["payload"]["text"] for res in search_results], search_results
        else:
            raise("Not implemented")
        
    async def set_current_collection(self, collection_name: str):
        self.db_client.set_collection(collection_name)
        return await self.db_client.init_collection()
    
    @abstractmethod
    async def add_doc(self, path: str) -> None:
        pass

    @abstractmethod
    async def delete_doc(self, filename: str, conditions) -> None:
        pass

    async def dump_from_folder(self, dir: str, user_chat_id: str) -> None:
        """
        Add documents from a directory to the database.

        Args:
            dir (str): The directory path containing the documents.
            unstructed_client (MyUnstructuredClient): The client for unstructured data.
            embedding_client (EmbeddingClient): The client for embedding data.
            db_client (MyQdrantClient): The client for the database.

        Returns:
            None: This function does not return anything.

        This function iterates through the files in the specified directory, extracts data from each file,
        and adds the document information to the database using the provided clients.
        """

        assert os.path.isdir(dir), "Must be a valid directory"

        for path, dirnames, filenames in os.walk(dir):
            for filename in filenames:
                await self.add_doc(os.path.join(path, filename), user_chat_id)
    
    async def list_s3_files(self, session, bucket_name, prefix):
        async with session.create_client('s3', region_name="ap-southeast-1") as s3_client:
            paginator = s3_client.get_paginator('list_objects_v2')
            operation_parameters = {'Bucket': bucket_name, 'Prefix': prefix}
            page_iterator = paginator.paginate(**operation_parameters)
            
            files = []
            async for page in page_iterator:
                if "Contents" in page:
                    for item in page["Contents"]:
                        files.append(item["Key"])
        return files
    
    async def download_file(self, session, bucket_name, key, download_dir):
        async with session.create_client('s3', region_name="ap-southeast-1") as s3_client:
            try:
                response = await s3_client.get_object(Bucket=bucket_name, Key=key)
                file_path = os.path.join(download_dir, key)
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
                async with response['Body'] as stream:
                    with open(file_path, 'wb') as f:
                        chunk = await stream.read()
                        f.write(chunk)
                return file_path
            except Exception as e:
                return f"Failed to download {key}: {e}"
            
    async def dump_from_s3_folder(self, dir: str, user_chat_id: str) -> None:

        session = get_session()
        files_to_download = await self.list_s3_files(session, os.environ.get("S3_BUCKET_NAME"), dir)
        if len(files_to_download) == 0:
            raise Exception("No files found in S3 bucket")
        
        tasks = [
            self.download_file(session, os.environ.get("S3_BUCKET_NAME"), key, os.environ.get("S3_DOWNLOAD_DIR"))
            for key in files_to_download
        ]
        results = await asyncio.gather(*tasks)
        failed = [result for result in results if "Failed to download" in result]
        results = [result for result in results if "Failed to download" not in result]

        if not isinstance(self.embedding_client, S3EmbeddingClient):
            for filename in results: #TODO: prevent processing of duplicate files
                await self.add_doc(filename, user_chat_id)
            
            ### When not using azure free tier, can use this instead of the above, maybe set max concurrency to <15
            # tasks = [self.add_doc(filename) for filename in results] #TODO: prevent processing of duplicate files
            # _ = await asyncio.gather(*tasks)
        else:
            # has to be done in batches so need a new method
            await self.add_doc_from_s3(results, user_chat_id)

        return failed

    async def dump_from_s3(self, file_list: List[str], user_chat_id: str, is_chunked: bool) -> None:

        session = get_session()
        tasks = [
            self.download_file(session, os.environ.get("S3_BUCKET_NAME"), key, os.environ.get("S3_DOWNLOAD_DIR"))
            for key in file_list
        ]
        print("starting to download")
        results = await asyncio.gather(*tasks)
        download_failed = [result for result in results if "Failed to download" in result]
        download_success = [result for result in results if "Failed to download" not in result]
        
        print("starting to process")
        success, char_counts = [], []

        for file in download_success:
            path, char_count = await self.add_doc_from_s3(file, user_chat_id, is_chunked)
            char_counts.append(char_count)
            success.append(path)

        return success, char_counts


from typing import Dict, List, AsyncIterator, TypedDict, Literal, Optional, Any
import time
import json
import logging

from openai import AsyncOpenAI

from animuz_core.genai.base import BaseAgent
from animuz_core.genai.tools import OpenAITool, MCPTools

logger = logging.getLogger(__name__)


# Type definitions for output structure
class MessageOutput(TypedDict):
    type: Literal["message"]
    role: str
    content: str

class FunctionCallOutput(TypedDict):
    type: Literal["function_call"]
    name: str
    arguments: str
    call_id: str

class FunctionCallResultOutput(TypedDict, total=False):
    type: Literal["function_call_output"]
    call_id: str
    output: str
    metadata: Any  # Filtered out before sending to OpenAI

OutputItem = MessageOutput | FunctionCallOutput | FunctionCallResultOutput


class OpenAIAgentClientStreaming(BaseAgent):
    """Streaming OpenAI Agent Client that yields SSE-compatible events.

    Supports both local tool dispatch and MCP-based tool execution.
    Uses the OpenAI Responses API for streaming.

    Args:
        user_chat_id: User/tenant ID for multi-tenant tool calls
        tools: Local tool dict (name -> (schema, callable))
        model: OpenAI model name (default: gpt-4o-mini)
        max_token: Max output tokens
        mcp_url: MCP server URL. If provided, tools are fetched from MCP.
        user_context: User context dict passed to MCP tool executions
        mcp_api_key: API key for MCP server. Defaults to MCP_API_KEY env var.
    """
    def __init__(
        self,
        user_chat_id: str,
        tools: Dict[str, tuple] = None,
        model: str = None,
        max_token: int = 2000,
        mcp_url: str = None,
        user_context: Dict = None,
        mcp_api_key: str = None,
        profiler=None,
    ) -> None:
        self.client = AsyncOpenAI()
        self.temperature = 0.0
        self.model = model if model else "gpt-4o-mini"
        self.max_token = max_token
        self.profiler = profiler
        self.mcp = None
        if mcp_url:
            self.mcp = MCPTools(mcp_url, user_chat_id, user_context, mcp_api_key, profiler=profiler)
        elif tools:
            self.add_tools(tools)

    def add_tools(self, tools: Dict[str, tuple]):
        self.tools = tools
        self.tool_dump = [
            {"type": "function", "function": tool[0].model_dump()}
            for tool in self.tools.values()
        ]

    def _filter_for_openai(self, messages: List[Dict]) -> List[Dict]:
        """Remove non-OpenAI fields (metadata, complete) before sending to API."""
        clean_messages = []
        for msg in messages:
            clean_msg = msg.copy()
            clean_msg.pop("metadata", None)
            clean_msg.pop("complete", None)

            if clean_msg.get("type") == "message":
                clean_messages.append({
                    "type": "message",
                    "role": clean_msg.get("role"),
                    "content": clean_msg.get("content")
                })
            elif clean_msg.get("type") == "function_call":
                clean_messages.append({
                    "type": "function_call",
                    "name": clean_msg.get("name"),
                    "arguments": clean_msg.get("arguments"),
                    "call_id": clean_msg.get("call_id")
                })
            elif clean_msg.get("type") == "function_call_output":
                clean_messages.append({
                    "type": "function_call_output",
                    "call_id": clean_msg.get("call_id"),
                    "output": clean_msg.get("output")
                })
            else:
                clean_messages.append(clean_msg)

        return clean_messages

    async def stream_reply_frontend(
        self,
        messages: List[Dict[str, str]],
        system_prompt: str = None
    ) -> AsyncIterator[dict]:
        """Stream reply events for frontend consumption.

        Yields dict events in two categories:

        STREAMING EVENTS (for real-time display):
            - {"type": "content_chunk", "content": "..."}
            - {"type": "tool_status", "message": "...", "function_name": "..."}
            - {"type": "tool_executing", "message": "...", "function_name": "...",
               "function_call_id": "...", "function_arguments": {...}}

        COMPLETE EVENTS (for backend/DB storage, marked with "complete": True):
            - {"type": "message", "role": "assistant", "content": "...", "complete": True}
            - {"type": "function_call", "name": "...", "arguments": "...", "call_id": "...", "complete": True}
            - {"type": "function_call_output", "call_id": "...", "output": "...", "complete": True}
        """
        messages_with_prompt = [{"role": "system", "content": system_prompt}] + messages
        user_message = messages[-1]["content"]

        output: List[OutputItem] = []

        async for event in self._create_message_streaming(
            self._filter_for_openai(messages_with_prompt),
            output
        ):
            yield event

        # Tool calling loop
        while self._has_function_calls_in_output(output):
            function_calls = self._extract_function_calls_from_output(output)

            for func_call in function_calls:
                function_name = func_call.get("name")
                args = json.loads(func_call.get("arguments", "{}"))
                call_id = func_call.get("call_id")

                logger.info(f"CALLING FUNCTION: {function_name} WITH ARGS: {str(args)}")

                yield {
                    "type": "tool_executing",
                    "message": f"Executing {function_name}...",
                    "function_name": function_name,
                    "function_call_id": call_id,
                    "function_arguments": args
                }

                result, metadata = await self.process_tool_use_mcp(function_name, args, user_message)

                tool_result: FunctionCallResultOutput = {
                    "type": "function_call_output",
                    "call_id": call_id,
                    "output": result,
                    "metadata": metadata
                }
                output.append(tool_result)
                messages_with_prompt.append(tool_result)

                yield {**tool_result, "complete": True}

            async for event in self._create_message_streaming(
                self._filter_for_openai(messages_with_prompt),
                output
            ):
                yield event

    async def _create_message_streaming(
        self,
        history: List[Dict],
        output: List[OutputItem]
    ) -> AsyncIterator[dict]:
        """Stream a single OpenAI Responses API call and accumulate output."""
        if self.profiler:
            self.profiler.mark("openai_call_start")
        start = time.time()

        # GPT-5 models don't support temperature
        create_kwargs = dict(
            model=self.model,
            max_output_tokens=self.max_token,
            input=history,
            tools=self.mcp.tools if self.mcp else getattr(self, 'tool_dump', []),
            stream=True
        )
        if self.model not in ["gpt-5", "gpt-5-mini", "gpt-5-nano"]:
            create_kwargs["temperature"] = self.temperature

        stream = await self.client.responses.create(**create_kwargs)

        accumulated_content = ""
        final_tool_calls = {}

        async for chunk in stream:
            if chunk.type == "response.in_progress":
                yield {"type": "in_progress"}
            elif chunk.type == "response.completed":
                yield {"type": "completed"}
            elif chunk.type == "response.error":
                raise Exception(f"OpenAI streaming error: {chunk.error.message}")

            if chunk.type == "response.output_text.delta":
                text_chunk = chunk.delta
                accumulated_content += text_chunk
                yield {"type": "content_chunk", "content": text_chunk}

            elif chunk.type == "response.output_item.added":
                if chunk.item.type == "function_call":
                    index = chunk.output_index
                    final_tool_calls[index] = {
                        "function_name": chunk.item.name,
                        "function_call_id": chunk.item.call_id,
                        "arguments": ""
                    }
                    yield {
                        "type": "tool_status",
                        "message": f"Preparing to use {chunk.item.name}...",
                        "function_name": chunk.item.name,
                        "function_call_id": chunk.item.call_id
                    }

            elif chunk.type == "response.function_call_arguments.delta":
                index = chunk.output_index
                if index in final_tool_calls:
                    final_tool_calls[index]["arguments"] += chunk.delta

        openai_duration = time.time() - start
        logger.info(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "OpenAI API latency (streaming)",
            "latency": round(openai_duration, 4),
        }))
        if self.profiler:
            self.profiler.mark("openai_call_end")
            self.profiler.record_duration("openai_api", openai_duration)

        if accumulated_content:
            message_output: MessageOutput = {
                "type": "message",
                "role": "assistant",
                "content": accumulated_content
            }
            output.append(message_output)
            yield {**message_output, "complete": True}

        for call in final_tool_calls.values():
            function_call_output: FunctionCallOutput = {
                "type": "function_call",
                "name": call["function_name"],
                "arguments": call["arguments"],
                "call_id": call["function_call_id"]
            }
            output.append(function_call_output)
            yield {**function_call_output, "complete": True}

    async def process_tool_use_mcp(self, tool_name: str, tool_input: dict, user_message: str):
        """Execute tool via MCP and return (result, metadata)."""
        start = time.time()
        try:
            response = self.mcp.execute_tool(tool_name, user_message=user_message, **tool_input)
            result, metadata = self.mcp.parse_response(tool_name, response, tool_input)

            logger.info(json.dumps({
                "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
                "message": f"Tool {tool_name} used successfully",
                "latency": round(time.time() - start, 4),
            }, ensure_ascii=False))

            return result, metadata
        except Exception as e:
            logger.error(f"Tool {tool_name} error: {str(e)}")
            return "ERROR", None

    def _has_function_calls_in_output(self, output: List[OutputItem]) -> bool:
        """Check if the last batch of output items contains function calls."""
        if not output:
            return False
        for item in reversed(output):
            if item.get("type") == "function_call":
                return True
            elif item.get("type") == "function_call_output":
                return False
        return False

    def _extract_function_calls_from_output(self, output: List[OutputItem]) -> List[Dict[str, str]]:
        """Extract function calls from the last batch of output items."""
        if not output:
            return []
        last_batch = []
        for item in reversed(output):
            if item.get("type") == "function_call_output":
                break
            last_batch.insert(0, item)
        return [
            {"name": item.get("name", ""), "arguments": item.get("arguments", "{}"), "call_id": item.get("call_id", "")}
            for item in last_batch
            if item.get("type") == "function_call"
        ]

    async def get_reply_frontend(self, messages: List[Dict[str, str]], system_prompt: str = None) -> str:
        raise NotImplementedError("This client is streaming-only. Use stream_reply_frontend instead.")



"""Pipelines module - RAG pipeline implementations."""


def __getattr__(name):
    if name == "BasePipeline":
        from .base import BasePipeline
        return BasePipeline
    if name == "AgenticRAG":
        from .agentic_rag import AgenticRAG
        return AgenticRAG
    if name == "SimpleRAG":
        from .simple_rag import SimpleRAG
        return SimpleRAG
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "BasePipeline",
    "SimpleRAG",
    "AgenticRAG",
]



import hashlib
import json, uuid, time, os
from typing import Dict, List, Tuple

from qdrant_client import models, AsyncQdrantClient

from animuz_core.utils import init_logger

LOGGER = init_logger(__name__)

def merge_result(result: List[List[models.ScoredPoint]]) -> List[dict]:
    """
        combine dense and sparse search results by summing the scores (following the bge-m3 paper)
        the results must be [[dense_scores,...], [sparse_scores,...]]
    """
    assert len(result) == 2, "the results must be [[dense_scores,...], [sparse_scores,...]]"

    dense = {p.id:p for p in result[0]}
    sparse = {p.id:p for p in result[1]}

    new_result = []
    for k,v in sparse.items():
        if k in dense:
            v = v.model_copy()
            v.score += dense[k].score
            dense.pop(k)

        new_result.append(v)

    for k,v in dense.items(): new_result.append(v)
        
    sorted_res = sorted(new_result, key=lambda x: x.score, reverse=True)
    return [res.dict() for res in sorted_res]

class MyQdrantClient():

    def __init__(self, host:str, port:int, collection_name:str) -> None:
        self.HOST = host
        self.PORT = port
        self.COLLECTION_NAME = collection_name

        self.client = AsyncQdrantClient(host=self.HOST, port= self.PORT, api_key=os.environ.get("QDRANT_CLOUD_API_KEY"))
        
    
    # async def init_collection(self) -> str:
    #     """
    #     Initializes a collection in the database if it does not already exist.
    #     If the collection does not exist, it creates a new collection with the specified parameters.
    #     Returns a string indicating whether the collection was created or if it already exists.
    #     """
    #     if not await self.client.collection_exists(self.COLLECTION_NAME):
    #         await self.client.create_collection(
    #             collection_name=self.COLLECTION_NAME,
    #             vectors_config={
    #                 "dense": models.VectorParams(
    #                     size=1024,
    #                     distance=models.Distance.DOT,
    #                 )
    #             },
    #             sparse_vectors_config={
    #                 "sparse": models.SparseVectorParams(index=models.SparseIndexParams(on_disk=True,)),
    #             }
    #         )
    #         return "collection created"
        
    #     return "collection exists"
    
    def set_collection(self, collection_name: str) -> None:
        self.COLLECTION_NAME = collection_name

    async def upload(self, dense_embedding: List[float], indices: List[str], values: List[float], properties: Dict[str, str]) -> None:
        """
            Upload the given dense_embedding and sparse embedding (in the form of `indices` and `values`) and a dict of properties to the vector db

        """
        start = time.time()
        self.client.upload_points(
            collection_name=self.COLLECTION_NAME,
            points=[
                models.PointStruct(
                    id=str(uuid.uuid4()),
                    vector={
                            "dense" : dense_embedding,
                            "sparse" : models.SparseVector(indices=indices, values=values),
                            },
                    payload=properties
                )
            ],
        )
        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Qdrant upload latency",
            "latency": round(time.time() - start, 4),
            "exception": False
        }))

    async def upload_batch(self, dense_embeddings: List[List[float]], list_of_indices: List[List[str]], list_of_values: List[List[float]], list_of_properties: List[Dict[str, str]]) -> None:
        """
            Batch upload the given dense_embeddings and sparse embeddings (in the form of `indices` and `values`) and dicts of properties to the vector db
            
        """
        start = time.time()
        points = [
            models.PointStruct(
                    id=str(uuid.uuid4()), 
                    vector={
                            "dense" : dense_embedding,
                            "sparse" : models.SparseVector(indices=indices, values=values),
                            }, 
                    payload=properties
            ) for dense_embedding, indices, values, properties in zip(dense_embeddings, list_of_indices, list_of_values, list_of_properties)
        ]

        self.client.upload_points(
            collection_name=self.COLLECTION_NAME,
            points=points,
        )
        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Qdrant batch upload latency",
            "latency": round(time.time() - start, 4),
            "exception": False
        }))

    async def upload_batch_v2(
        self,
        dense_embeddings: List[List[float]],
        list_of_sparse_vectors: List[Dict[str, Dict[str, List]]],
        list_of_properties: List[Dict[str, str]]
    ) -> None:
        """
        Batch upload the given dense_embeddings and multiple sparse embeddings (e.g., 'sparse', 'bm25') and dicts of properties to the vector db.
        list_of_sparse_vectors: List of dicts, each mapping sparse vector name to {"indices": [...], "values": [...]}

        This is better for cases where we have multiple sparse embeddings per point.
        We also generate point IDs here based on the properties.
        """
        start = time.time()
        points = []
        for dense_embedding, sparse_vectors, properties in zip(dense_embeddings, list_of_sparse_vectors, list_of_properties):
            vector_dict = {"dense": dense_embedding}
            for name, vec in sparse_vectors.items():
                vector_dict[name] = models.SparseVector(indices=vec["indices"], values=vec["values"])
            
            # Point ID Logic
            if "point_id" in properties:
                point_id = properties["point_id"]
                # Remove point_id from properties so it is not included in the payload
                properties.pop("point_id", None)
            else:
                # Use md5 as uuid (128 bits / 16 bytes / 32 hex chars)
                prop_str = json.dumps(properties, sort_keys=True)
                point_id = hashlib.md5(prop_str.encode("utf-8")).hexdigest()

            points.append(
                models.PointStruct(
                    id=point_id, 
                    vector=vector_dict,
                    payload=properties
                )
            )

        self.client.upload_points(
            collection_name=self.COLLECTION_NAME,
            points=points,
        )
        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Qdrant batch upload latency",
            "latency": round(time.time() - start, 4),
            "exception": False
        }))

    async def hybrid_search(self, dense_embedding: List[float], indices: List[str], values: List[float], user_chat_id: str, top_k: int = 3):
        """Hybrid search using server-side RRF fusion (single round trip)."""
        start = time.time()
        limit = 10 if top_k == 3 else 20

        result = await self.client.query_points(
            collection_name=self.COLLECTION_NAME,
            prefetch=[
                models.Prefetch(
                    query=models.SparseVector(indices=indices, values=values),
                    using="sparse",
                    limit=limit,
                ),
                models.Prefetch(
                    query=dense_embedding,
                    using="dense",
                    limit=limit,
                ),
            ],
            query_filter=models.Filter(
                must=[
                    models.FieldCondition(
                        key="userChatID",
                        match=models.MatchValue(value=user_chat_id),
                    )
                ]
            ),
            query=models.FusionQuery(fusion=models.Fusion.RRF),
            limit=top_k,
        )

        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Hybrid search latency",
            "latency": round(time.time() - start, 4),
            "exception": False
        }))
        return result

    async def delete_doc(self, conditions: List[Tuple[str, str]]) -> None:
        start = time.time()
        await self.client.delete(
            collection_name=self.COLLECTION_NAME,
            points_selector=models.FilterSelector(
                filter=models.Filter(
                    must=[
                        models.FieldCondition(
                            key=cond[0],
                            match=models.MatchValue(value=cond[1]),
                        ) for cond in conditions
                    ],
                )
            ),
        )
        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Qdrant delete doc latency",
            "latency": round(time.time() - start, 4),
            "exception": False
        }))


from typing import List, Dict
import logging

# from src.genai.ollama_client import OLlamaClient
# from src.genai.anthropic_client import AnthropicClient

from pydantic import BaseModel

class ChatRequest(BaseModel):
    messages: List[Dict[str, str]]
    user_chat_id: str

# async def extract_metadata(chunks: List[str], model_name="llama3-8b") -> Dict[str, str]:
#     # client = None
#     client = AnthropicClient()

#     return await client.summarize_chunks(chunks)

def format_sources(sources):
    """
    Takes a list of sources and returns a formatted numbered list as a string.
    
    Args:
    sources (list): A list of source strings.

    Returns:
    str: A formatted numbered list of sources.
    """
    formatted_sources = ""
    for idx, source in enumerate(sources, 1):
        formatted_sources += f"Source {idx}. {source}\n"
    return formatted_sources.strip()

def init_logger(name):
    LOGGER = logging.getLogger(name)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(filename)s - %(module)s: %(funcName)s: %(lineno)d - %(message)s')
    LOGGER.setLevel(logging.DEBUG)
    try:
        import watchtower
        cw_handler = watchtower.CloudWatchLogHandler(log_group="animuz", stream_name="test", use_queues=False)
        cw_handler.setFormatter(formatter)
        LOGGER.addHandler(cw_handler)
    except Exception:
        pass
    return LOGGER

if __name__ == "__main__":
    pass
    # import asyncio
    # from dotenv import load_dotenv

    # unstructured_client = MyUnstructuredClient(host="localhost", port=12083)
    # load_dotenv()
    # azure_doc_ai_key: str = os.environ.get("AZURE_DOC_AI_KEY")
    # azure_doc_ai_endpoint: str = os.environ.get("AZURE_DOC_AI_ENDPOINT")
    # # azure_doc_ai_client = AzureDocAiClient(endpoint=azure_doc_ai_endpoint, key=azure_doc_ai_key)
    # import pickle
    # async def test(path):
    #     # res = await azure_doc_ai_client.analyze_document(path)
    #     with open("temp.pkl", "rb") as f:
    #         res = pickle.load(f)

    #     return await extract_metadata([str(data_row["text"]) for data_row in res])
    # res = asyncio.run(test("test_docs/test.pdf"))
    # print(res)


import asyncio
from typing import List
import ollama

from animuz_core.genai.base import BaseLLM
from animuz_core.genai.prompts import SUMMARIZE_PROMPT, COMBINE_SUMMARIES_PROMPT

class OLlamaClient(BaseLLM):
    def __init__(self, system_prompt: str = None) -> None:
        self.system_prompt = system_prompt
        self.client = ollama.AsyncClient(host="http://127.0.0.1:12084")
        self.message_history = []
        self.temperature = 0.0

    async def get_reply(self, text: str):
        """
        Appends the user's input text to the message history, creates a message using the client with specified parameters, 
        and appends the assistant's response to the message history. Returns the assistant's response text.
        
        Parameters:
            text (str): The user's input text.
        
        Returns:
            str: The assistant's response text.
        """
        self.message_history.append({"role": "user", "content": text})

        response = await self.client.chat(model='llama3', messages=self.message_history)

        self.message_history.append({"role": "assistant", "content": response['message']['content']})
        return response['message']['content']
    
    async def summarize(self, text:str):

        messages = [
            {"role": "system", "content" : SUMMARIZE_PROMPT},
            {"role": "user", "content": text}
                    ]
        response = await self.client.chat(model='llama3:70b', messages=messages)
        print(response)
        print()
        return response['message']['content']
    
    async def summarize_chunks(self, chunks: List[str]):
        res = await asyncio.gather(*[self.summarize(chunk) for chunk in chunks])

        messages = [
            {"role": "system", "content" : COMBINE_SUMMARIES_PROMPT},
            {"role": "user", "content": "\n".join(res)}
                    ]
        response = await self.client.chat(model='llama3:70b', messages=messages)

        return response['message']['content']



"""Embedding module - Multiple embedding client implementations."""

from .embedding_client import HTTPEmbeddingClient, EmbeddingClient
from .modal_embedding_client import ModalEmbeddingClient
from .s3_embedding_client import S3EmbeddingClient

__all__ = [
    "HTTPEmbeddingClient",
    "EmbeddingClient",
    "ModalEmbeddingClient",
    "S3EmbeddingClient",
]



import asyncio, json
from typing import List, Dict
from pathlib import Path

from animuz_core.pipelines.base import BasePipeline
from animuz_core.genai.base import BaseAgent
from animuz_core.embedding.embedding_client import EmbeddingClient
from animuz_core.ingest.custom_unstructured_client import MyUnstructuredClient
from animuz_core.ingest.azure_doc_ai_client import AzureDocAiClient
from animuz_core.vectordb.utils import MyQdrantClient

from animuz_core.genai.tools import AnthropicRetrieverTool, OpenAIRetrieverTool

class AgenticRAG(BasePipeline):
    def __init__(self, embedding_client: EmbeddingClient, db_client: MyQdrantClient, agent: BaseAgent, tools: List[str] = [], unstructured_client: MyUnstructuredClient = None, azure_doc_ai_client: AzureDocAiClient = None) -> None:
        """
        Initializes the AgenticRAG object with the specified clients and models.

        Args:
            embedding_client (EmbeddingClient): The client for embedding data.
            db_client (MyQdrantClient): The client for the database.
            agent (BaseAgent): The LLM agent for the AgenticRAG object.
            tools (List[str]): The list of tool names available for the LLM agent.
            unstructured_client (MyUnstructuredClient, optional): The client for unstructured data. If None, uses the default MyUnstructuredClient.
            azure_doc_ai_client (AzureDocAiClient, optional): The client for Azure Doc AI. If None, uses the default AzureDocAiClient.

        Returns:
            None: This function initializes the AgenticRAG object without returning anything.
        """
        super().__init__(embedding_client, db_client, unstructured_client, azure_doc_ai_client)

        self.agent = agent
        if tools:
            self.tools = self.create_tools(tools)
            self.agent.add_tools(self.tools)

    def create_tools(self, tool_names: List[str]):
        tools = dict()
        for tool_name in tool_names:
            if tool_name == "anthropic_retriever":
                tools["retriever"] = (AnthropicRetrieverTool.create_tool(), self.query_db)
            elif tool_name == "openai_retriever":
                tools["retriever"] = (OpenAIRetrieverTool.create_tool(), self.query_db)
        return tools

    async def add_doc(self, path: str, user_chat_id: str, is_chunked: bool) -> None:
        """
        Add a document to the database.

        Args:
            path (str): The path to the document.

        Returns:
            None: This function does not return anything.

        Raises:
            None: This function does not raise any exceptions.

        This function adds a document to the database by performing the following steps:
        1. Partition the document using the unstructured client or azure doc ai.
        2. Extract the properties of each partitioned data row.
        3. Get the embeddings for each partitioned data row using the embedding client.
        4. Extract the indices and values from the sparse embeddings.
        5. Upload the dense and sparse embeddings, indices, values, and properties to the database using the db client.

        Note: The document is expected to be in the specified path.
        """
        print("START parsing for doc", path)
        res = await self.parse_doc_markdown(path, is_chunked)
        print("done parsing for doc", path)
        properties = [{
                "text": data_row["text"],
                "source": data_row["source"],
                "fileDirectory": str(Path(path).parent),
                "filename": str(Path(data_row["metadata"]["filename"]).stem),
                "filetype": Path(data_row["metadata"]["filename"]).suffix,
                "userChatID": user_chat_id
            } for data_row in res]
        
        if len(res) < 1000:
            semaphore = asyncio.Semaphore(10)
            tasks = []
            emb_res = []
            for data_row in res:
                async with semaphore:
                    tasks.append(self.embed_single(str(data_row["text"])))
                    await asyncio.sleep(1)

            for task in tasks:
                emb_res.append(await task)

        else:
            emb_res = await self.batch_embed([str(data_row["text"]) for data_row in res])

        print("done embedding for doc", path)
        indices = [e["sparse_embedding"].keys() for e in emb_res]
        values = [e["sparse_embedding"].values() for e in emb_res]
            
        await self.upload_to_db([e["embedding"] for e in emb_res], indices, values, properties)
        print("done uploading for doc", path)

    async def add_doc_from_s3(self, paths: List[str], user_chat_id: str, is_chunked: bool) -> None:
        semaphore = asyncio.Semaphore(10)
        tasks = []
        res = []
        for path in paths:
            async with semaphore:
                tasks.append(self.parse_doc_markdown(path, is_chunked))
                await asyncio.sleep(1)

        for task in tasks:
            res.extend(await task)

        properties = [{
            "text": data_row["text"],
            "source": data_row["source"],
            "fileDirectory": str(Path(path).parent),
            "filename": str(Path(data_row["metadata"]["filename"]).stem),
            "filetype": Path(data_row["metadata"]["filename"]).suffix,
            "userChatID": user_chat_id
        } for data_row in res]

        emb_res = await self.batch_embed([str(data_row["text"]) for data_row in res])

        indices = [e["sparse_embedding"].keys() for e in emb_res]
        values = [e["sparse_embedding"].values() for e in emb_res]
            
        await self.upload_to_db([e["embedding"] for e in emb_res], indices, values, properties)

    async def delete_doc(self, filename: str, user_chat_id: str) -> None:
        conditions = [("fileName", str(Path(filename).stem)), 
                      ("fileType", Path(filename).suffix),
                      ("userChatID", user_chat_id)]
        await self.db_client.delete_doc(conditions= conditions)

    async def chat(self, messages: List[Dict[str, str]], user_chat_id: str) -> str:
        return await self.agent.get_reply_frontend(messages, user_chat_id)
    
    async def chat_two(self, messages: List[Dict[str, str]], user_chat_id: str, background_tasks) -> str:
        return await self.agent.get_reply_frontend_tool_use_ack(messages, user_chat_id, background_tasks)
    
    async def retrieve_result(self, chat_id: str) -> str:
        return await self.agent.get_chat_result(chat_id)
    
    async def streaming_chat(self, messages: List[Dict[str, str]], user_chat_id: str):
        async for chunk in self.agent.get_reply_frontend_streaming(messages, user_chat_id):
            yield json.dumps(chunk, ensure_ascii=False) + "\n"

    def clear_history(self) -> None:
        self.agent.clear_history()


CLAUDE_RAG_DEFAULT_PROMPT="""Given the following conversation, relevant context, and a follow up question, reply with an answer to the current question the user is asking. 
Return only your response to the question given the above information following the users instructions as needed."""

CLAUDE_TOOL_DEFAULT_PROMPT="""You are an AI assistant for KP trading group with access to a vector database containing knowledge, e.g. employee handbook, of the KP group. 
Users (which are mostly internal employees and executives) will mostly ask questions related to the KP group. So it is safe to assume that the queries are related to the KP group.
However, if the queries pertains to a general question, answer using your own knowledge first. Only query the vector database if you need additional context to fully answer.
When a user asks a query:

1. Attempt to provide a complete answer using your current knowledge.
2. If you cannot fully answer, query the vector database for relevant supplementary information.
3. Generate a final answer combining your knowledge and any retrieved context.
4. Provide the answer, citing vector database sources if used.

Be concise. Only use the vector database when truly needed to comprehensively respond to the query. When creating a query, ensure that it is comprehensive and in Thai.
"""

SUMMARIZE_PROMPT = """You are a skilled document summarizer. Your task is to summarize individual chunks of a larger document in a concise and informative manner.

For each chunk of text provided, you will:

1. Read and understand the content of the chunk.
2. Identify the main ideas, key points, and important information present in the chunk.
3. Generate a concise summary of the chunk, focusing on the most essential information while omitting unnecessary details.
4. Ensure that the summary is clear, concise, coherent, and accurately represents the content of the chunk.
5. Ensure that the summary is in THAI. THIS IS VERY IMPORTANT

When summarizing each chunk, keep in mind that the individual summaries will eventually be combined into a single summary of the entire document. Therefore, it's important to maintain a consistent tone, style, and level of detail across all chunk summaries."""

COMBINE_SUMMARIES_PROMPT = """You have been provided with summaries for different sections of a document. Combine these chunk summaries into one cohesive summary representing the entire document.

Steps:
1. Review all chunk summaries to understand their key points and information.
2. Note any overlaps, redundancies, or inconsistencies across summaries.
3. Determine a logical structure and flow for the final summary.
4. Combine the summaries into one cohesive piece:
   - Maintain consistent tone, style, and detail level
   - Eliminate redundancies
   - Smoothly transition between sections
   - Preserve essential meaning and key points
5. Refine and polish the combined summary for clarity and conciseness.
6. Ensure that the summary is in THAI.

Produce a well-organized, accurate summary that effectively communicates the document's content without introducing new information. Prioritize clarity, coherence, and representing the original text faithfully."""


RUDI_DEFAULT_PROMPT="""You are a witty and helpful chatbot named Rudi for K.P. Trading Group. Rudi is a chatdog developed by Animuz. K.P. Trading Group is the sole distributor for Hosiwell's cables. You are different than other chatbots because you respond just like how a polite person would perform text messaging, separating the message into multiple parts with double slash // and occasionally using emojis to make the conversation more engaging. Users will be outside customers arriving on the web. You are expected to provide information about the products and services of K.P. Trading Group, as well as general information about the company. However, if the queries pertains to a general question, answer using your own knowledge first. Only query the vector database if you need additional context to fully answer.
When a user asks a query:

1. Attempt to provide a complete answer using your current knowledge.
2. If you cannot fully answer, query the vector database in Thai for relevant supplementary information.
3. Generate a final answer combining your knowledge and any retrieved context.
4. Provide the answer, citing vector database sources if used.

Only use the vector database when truly needed to comprehensively respond to the query. Only if you don't have the answer to the question, respond with K.P's contact information for the user to contact admin."""
